/*:ts=8*/
/*****************************************************************************
 * FIDOGATE --- Gateway UNIX Mail/News <-> FTN NetMail/EchoMail
 *
 * $Id: route.c,v 1.2 2002/04/28 21:32:47 dyff Exp $
 *
 * Making dynamic routing
 *
 *****************************************************************************
 * Copyright (C) 2001
 * 
 *    Dmitry Fedotov            FIDO:      2:5030/1229
 *				Internet:  dyff@fidogate.spb.ru
 *
 * This file is part of FIDOGATE.
 *
 * FIDOGATE is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * FIDOGATE is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with FIDOGATE; see the file COPYING.  If not, write to the Free
 * Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************/

#include <signal.h>

#include "fidogate.h"
#include "getopt.h"


#define PROGRAM 	"route"
#define VERSION 	"$Revision: 1.2 $"
#define CONFIG		DEFAULT_CONFIG_MAIN


static	char *route_file = NULL;
static	char *true_file  = NULL;
static	char *nl_file    = NULL;

static int signal_exit   = FALSE;	/* Flag: TRUE if signal received */

Nodelist  *nl_first       = NULL;
Nodelist  *nl_last        = NULL;



typedef struct st_statuslist
{
    char   *st;
    struct st_statuslist *next;
}Statuslist;

static Statuslist *status_list = NULL;
static Statuslist *status_last = NULL;

typedef struct st_nllist
{
    Statuslist *st;
    Statuslist *status_list;
    Statuslist *status_last;
    LON	        routing;
}NlList;

char *f_read(char *buffer, int len, FILE *fp)
{
    char *p;
    long cf_lineno = 0;

    while (fgets(buffer, len, fp)) {
	cf_lineno++;
	strip_crlf(buffer);
	for(p=buffer; *p && is_space(*p); p++) ; /* Skip white spaces */
	if (*p != ';')
	    return p;
    }
    return NULL;
}


#define	MODE_PVT	1
#define	MODE_HOLD	2
#define MODE_DOWN	3
#define	MODE_NORMAL	4
#define MODE_HUB	5
#define MODE_HOST	6
#define MODE_ZONE	7


void status_add(Statuslist *st, char *s)
{
    Statuslist *p;
    
    p = (Statuslist *)xmalloc(sizeof(Alias));
    p->st = s;
    p->next = NULL;

    if(status_list)
	status_last->next = p;
    else
	status_list = p;
    status_last = p;
}

void status_init(Statuslist *st)
{
    st->next = NULL;
    st->st = NULL;
}

/*
 * Scaning Nodelist
 */
void nl_scan ()
{
    LON rou;
    Node host, hub, node;
    FILE *fp;
    Statuslist st;
    char *p = NULL;
    char *s = NULL;
    char status;

    fp = fopen(nl_file, "r");
    if(!fp)
	debug(1, "ERROR: reading %s", nl_file);

    status_init(&st);

    while(f_read(buffer, sizeof(buffer), fp))
    {
	if(buffer[0] == ',')
	{
	    if( (p = xstrtok(buffer+1, ",")) )
	    {
		asc_to_node(p, &node, FALSE);
		status =  MODE_NORMAL;
	    }	    
	}
	else if( (p = xstrtok(buffer, ",")) && (s = xstrtok(NULL, ",")) )
	{

		if(stricmp(p, "Zone") )
		{
		    host.zone = atoi(s);
		    host.net  = 0;
		    host.node = 0;
		    status = MODE_ZONE;
		}
		else if(stricmp(p, "Host"))
		{
		    status = MODE_HOST;
		    if(!host.zone)
			host = cf_n_addr();
		    host.net  = atoi(s);
		    host.node = 0;
		}
		else if(stricmp(p, "Hub"))
		{
		    status = MODE_HUB;
		    asc_to_node(p, &hub, FALSE);
		}
		else if(stricmp(p, "Hold"))
		{
		    status = MODE_HOLD;
		}
		else if(stricmp(p, "Down"))
		{
		    status = MODE_DOWN;
		}
		else if(stricmp(p, "Pvt"))
		{
		    status = MODE_PVT;
		}
	}
	status_add(&st, &status);
	lon_add(&rou, &host);
    }
}

void trust_scan()
{


}

void route_scan()
{


}


/*
 * Function called on SIGINT
 */
void prog_signal(int signum)
{
    char *name = "";

    signal_exit = TRUE;
    
    switch(signum)
    {
    case SIGHUP:
	name = " by SIGHUP";  break;
    case SIGINT:
	name = " by SIGINT";  break;
    case SIGQUIT:
	name = " by SIGQUIT"; break;
    default:
	name = "";            break;
    }

    log("KILLED%s: exit forced", name);
}


void short_usage(void)
{
    fprintf(stderr, "usage: %s [-options]\n", PROGRAM);
    fprintf(stderr, "       %s --help  for more information\n", PROGRAM);
}


void usage(void)
{
    fprintf(stderr, "FIDOGATE %s  %s %s\n\n",
	    version_global(), PROGRAM, version_local(VERSION) );
    
    fprintf(stderr, "usage:   %s [-options]\n\n", PROGRAM);
    fprintf(stderr, "\
options: -g --grade G                 processing grade\n\
         -l --lock-file               create lock file while processing\n\
         -r --routing-file NAME       read routing file\n\
\n\
	 -v --verbose                 more verbose\n\
	 -h --help                    this help\n\
         -c --config NAME             read config file (\"\" = none)\n");

    exit(0);
}


int main (int argc, char **argv)
{
    char *r_flag=NULL, 
	 *c_flag=NULL;
    char *p;
    short int  l_flag = FALSE;
    int   c, ret;
    int   g_flag = FALSE;
    
    
    int option_index;
    static struct option long_options[] =
    {
	{ "grade",        1, 0, 'g'},	/* grade */
	{ "lock-file",    0, 0, 'l'},	/* Create lock file while processing */
	{ "routing-file", 1, 0, 'r'},	/* Set routing file */

	{ "verbose",      0, 0, 'v'},	/* More verbose */
	{ "help",         0, 0, 'h'},	/* Help */
	{ "config",       1, 0, 'c'},	/* Config file */
	{ 0,              0, 0, 0  }
    };

    log_program(PROGRAM);
    cf_initialize();
    
    
    while ((c = getopt_long(argc, argv, "g:lr:vhc:",
			    long_options, &option_index     )) != EOF)


	switch (c) {
	/***** ftnroute options *****/
	case 'g':
	    g_flag = *optarg;
	    break;
        case 'l':
	    l_flag = TRUE;
            break;
	case 'r':
	    r_flag = optarg;
	    break;
	    
	/***** Common options *****/
	case 'v':
	    verbose++;
	    break;
	case 'h':
	    usage();
	    exit(0);
	    break;
	case 'c':
	    c_flag = optarg;
	    break;
	default:
	    short_usage();
	    exit(EX_USAGE);
	    break;
	}

    cf_read_config_file(c_flag ? c_flag : CONFIG);
    cf_debug();
    
    signal(SIGHUP,  prog_signal);
    signal(SIGINT,  prog_signal);
    signal(SIGQUIT, prog_signal);

    ret = EXIT_OK;

    /* Lock file */
    if(l_flag)
        if(lock_program(PROGRAM, NOWAIT) == ERROR )
    	    /* Already busy */
    	    exit(EXIT_BUSY);


    if( (p = cf_get_string("RouteFile", TRUE)) )
    {
	debug(8, "config: RouteFile %s", p);
	route_file = p;
    }

    if( (p = cf_get_string("TrueFile", TRUE)) )
    {
	debug(8, "config: TrueFile %s", p);
	true_file = p;
    }

    if( (p = cf_get_string("NlFile", TRUE)) )
    {
	debug(8, "config: NlFile %s", p);
	nl_file = p;
    }
    else
    {
	log("ERROR: can't find nodelist file in global config");
	exit(1);
    }

    routing_init(r_flag ? r_flag : cf_p_routing() );

    nl_scan();

    if(l_flag)
        unlock_program(PROGRAM);


    exit(ret);
}
